(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program.
Require Import countable vgtac basis lang_common lang_spec.
Require Import model_common model_untyped model_properties_untyped gw_untyped mm.
Require Import ssr util opt_transform lang_mid lang_mid_renaming pilsner_common.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".


(** Commute let-bindings. *)


(** * Expression transformation. *)


(** ** Definition. *)

Definition commute (e : exp) : option exp :=
  match e with
    | IElet ea xa (IElet eb xb e') => 
      if list_mem eqdec (tkvar2var xa) (vars eb) then
        
        None
      else
        Some (IElet eb xb (IElet ea xa e'))
    | _ => None
  end.


(** ** Syntactic correctness. *)

Lemma commute_wfexp: forall e G e'
    (E: commute e = Some e')
    (WF: wfexp G e)
    (UN: uniq (bvars e ++ purge_lbl G)),
  <<WF: wfexp G e'>> /\
  <<BV: subuniq (bvars e') (bvars e)>> /\
  <<GT: is_good_thunk e -> is_good_thunk e'>> /\
  <<GE: good_exp e -> good_exp e'>>.
Proof.
  i; destruct e; vauto; destruct e2; vauto.
  destruct bx, bx0
  ; (s in E; des_ifs; s in WF; des; s; splits; try (by vauto);
     [ rewrite <-(app_nil_l G); eapply wfatom_strengthen; eauto
       ; eapply list_mem_false; eauto
     | rewrite <-(app_nil_l _); apply wfatom_weaken; eauto
     | eapply wfexp_exchange; eauto using perm_swap
     | apply subuniq_perm; rewrite !app_comm_cons 2!app_assoc
       ; eauto using Permutation_app, Permutation_app_comm]
    ).
Qed.



(** ** Semantic correctness. *)

Lemma commute_psim lw G ea eb: forall
    (CV: commute ea = Some eb)
    (WF: wfexp G ea)
    (UN: uniq (bvars ea ++ purge_lbl G))
    (SLW: stable lw),
  mm_psim lw G ren_id eb ea.
Proof.
  i; revert CV.
  destruct ea; try destruct ea2; try discriminate.
  destruct eb; try destruct eb2; unfold commute;
  case_eq (list_mem eqdec (tkvar2var bx) (vars ea2_1)); try discriminate.
  i; inversion CV; clarify; clear CV.
  rename eb2_1 into a, eb1 into b, bx1 into y, bx2 into x, eb2_2 into e.  
  rr; splits; vauto; []; exists 0; i.
  rename enva into env', envb into env.

  eapply pesim_acquire; i.
  eapply pesim_real; i.

  rrdes HR. dup WHX. apply mm_confrel_shape in WHX. rrdes WHX.
  subst; clarify; dup RLa; dup RLb.
  rrdes RLa; rrdes RLb; inversion Em; inversion Em0.
  destruct hay, hby; destruct e0, e1; inv Ec; inv Ec0; subst.
  clean. simpl cplus in *; noemp.
  
  assert (INxG: ~ In (tkvar2var x) G).
  { i; apply purge_lbl_tkvar2var in FH.
    eapply (uniq_appD _ _ UN); eauto.
    case x as [x|kx]; s; auto.
  }

  assert (INyG: ~ In (tkvar2var y) G).
  { i; apply purge_lbl_tkvar2var in FH.
    eapply (uniq_appD _ _ UN); eauto.
    case y as [y|ky], x as [x|kx]; s; right; eapply in_or_app; right; s; auto.
  }

  assert (uniq ([tkvar2var x] ++ [tkvar2var y])). 
  { rewrite <-app_nil_r at 1; eapply (uniq_sublist UN), sublist_app
    ; auto using sublist_nil'.
    case x, y; s bvars; eapply sublist_cons_keep
    ; rewrite <-app_nil_l at 1; eapply sublist_app; auto using sublist_nil'.
  }

  assert (x <> y).
  { rr; i; subst. s in H0. des. eapply H0. case y; i; s; auto. }

  assert (SIM: exists i', forall va va' vb vb'
                 (RVx: match x with
                         | inl _ => vclos (wlift lw) s (g s) va' va
                         | inr _ => ksim _ (wfo_nat i) kia kib g si s s va' va
                       end)
                 (RVy: match y with
                         | inl _ => vclos (wlift lw) s (g s) vb' vb
                         | inr _ => ksim _ (wfo_nat i) kia kib g si s s vb' vb
                       end),
   esim _ i' kia kib g si s false
     (ha' ## h1, Just ((tkvar2var x, va')::(tkvar2var y, vb')::env', e))
     (hb' ## h2, Just ((tkvar2var y, vb)::(tkvar2var x, va)::env, e))).
  { destruct (@mm_psim_refl lw e (tkvar2var y :: tkvar2var x :: G)) as [o' PSe]
    ; [by destruct y, x; vauto; invc WF H3|by auto|].
    des.
    econs; i; rewrite (mid_csplit (ha' ## _)) (mid_csplit (hb' ## _)).
    eapply pesim_release; i; try by rr; esplits; eauto.
    eapply PSe0; auto; i.
    - apply in_consD in IN; des.
      { destruct y; vauto; s in IN; clarify.
        rewrite list_assoc_cons_neq; [|by s in H0; rrdes H0; ii; eauto].
        esplits; eauto.
      }
      apply in_consD in IN; des.
      { destruct x; vauto; s in IN; clarify; simpl tkvar2var.
        rewrite list_assoc_cons_eq.
        rewrite list_assoc_cons_neq; [|by s in H0; rrdes H0; ii; eauto].
        esplits; eauto.
      }
      specialize (RV x0); rewrite ->ren_itvar_id in *. 
      repeat (rewrite list_assoc_cons_neq; [|congruence]).
      eapply RV; eauto.
    - apply in_consD in IN; des.
      { destruct y; vauto; s in IN; clarify.
        rewrite list_assoc_cons_neq; [|by s in H0; rrdes H0; ii; eauto].
        esplits; eauto.
      }
      apply in_consD in IN; des.
      { destruct x; vauto; s in IN; clarify; simpl tkvar2var.
        rewrite list_assoc_cons_eq.
        rewrite list_assoc_cons_neq; [|by s in H0; rrdes H0; ii; eauto].
        esplits; eauto.
      }
      specialize (RK x0); rewrite ->ren_contvar_id in *. 
      repeat (rewrite list_assoc_cons_neq; [|congruence]).
      eapply RK; eauto.
  }
  des.

  destruct x as [x|kx], y as [y|ky]; destruct WF as (WFa & WFb & WF); i.

  - (*| two regular lets *)

    destruct (LEM (exists va, eval env a = Some va)) as [Hva|Hva]; last first.
    { i. pfold. econs 1. i. destruct XTRa as [Xha], XTRb as [Xhb]; subst.
      econs 1.
      - econs 1. inv CNF; des; subst; rrdes MB; inv Ec.
        eapply IRerr; i; first by auto. inv FH. apply Hva. by eauto using EVAL.
      - eauto.
    }

    destruct Hva as [va Eva].
    
    destruct (LEM (exists vb, eval ((inl (inr x), va) :: env) b = Some vb)) as [Hvb|Hvb]; last first.
    { i. pfold. econs 1. i.
      destruct XTRa as [Xha XTRa], XTRb as [Xhb XTRb]. subst.
      econs 1.
      - inv CNF; des; subst; rrdes MB; inv Ec.
        econs 3; econs 1; inst_pairs.
        + econs 1. econs. by eauto.
        + econs 2; i; first by auto.
          inv FH. apply Hvb. by eauto using EVAL.
      - eauto.
    }
    destruct Hvb as [vb Evb].
    
    assert (Hvb': exists vb', eval env' b = Some vb'
                           /\ vclos (wlift lw) s (g s) vb' vb).
    { 
      eapply (mm_asim_refl _ G); [| | | |exact Evb]; eauto.
      - eapply (wfatom_strengthen _ nil); eauto; eapply list_mem_false; eauto.
      - i. destruct (RV x0); eauto; des; esplits; eauto.
        rewrite list_assoc_cons_neq; try by auto.
        rr; i; by clarify.
    }
    destruct Hvb' as [vb' [Evb' Vvb]].

    assert (Hva': exists va', eval ((inl (inr y), vb') :: env') a = Some va'
                           /\ vclos (wlift lw) s (g s) va' va).
    {
      eapply mm_asim_refl; [| | | |exact Eva]; eauto.
      i. destruct (RV x0); eauto; des; esplits; eauto.
      rewrite list_assoc_cons_neq; auto.
      rewrite ren_itvar_id; rr; i; by clarify.
    }    
    destruct Hva' as [va' [Eva' Vva]].

    eapply pesim_dtsteps.
    + eassumption.
    + econs 2; first by eapply mid_wdet.
      * inst_pairs. econs. econs. by eauto. 
      * econs 2; first by eapply mid_wdet.
        { inst_pairs. econs. econs. eauto. }
        { econs 1. }
    + i; econs. esplits; [| |exact PV| exact FIN]; inst_pairs. eauto. reflexivity.
    + eauto.
    + i; eexists; rrdes RLb. 
      destruct XTR as [Xh]; subst. inv Ec; clean.
      esplits; first 1 last. 
      * left; econs 2; econs 1; inst_pairs; econs 1; eapply IRlet; eauto.
      * rr; esplits; inst_pairs; s; noemp; eauto.
      * eapply SIM; eauto.

  - (*| regular let before continuation let *)

    destruct (LEM (exists va, eval env a = Some va)) as [Hva|Hva]; last first.
    { i. pfold. econs 1. i. destruct XTRa as [Xha], XTRb as [Xhb]; subst.
      econs 1.
      - econs 1. inv CNF; des; subst; rrdes MB; inv Ec.
        eapply IRerr; i; first by auto. inv FH. apply Hva. by eauto using EVAL.
      - eauto.
    }
    destruct Hva as [va Eva].

    destruct (LEM (exists bx e', b = IEcont bx e')) as [Hvb|Hvb]; last first.
    { i. pfold. econs 1. i.
      destruct XTRa as [Xha XTRa], XTRb as [Xhb XTRb]. subst.
      econs 1.
      - inv CNF; des; subst; rrdes MB; noemp; inv Ec.
        econs 3; econs 1; inst_pairs.
        + econs 1. econs. by eauto.
        + econs 2; i; first by auto.
          inv FH. apply Hvb. by eauto. 
      - eauto.
    }
    destruct Hvb as [bx [e']]; subst.
    s in WFb.

    assert (Hva': exists va', eval ((inr ky, IVclo env' (IEcont bx e')) :: env') a = Some va' 
                           /\ vclos (wlift lw) s (g s) va' va).
    { eapply mm_asim_refl; [| | | |exact Eva]; eauto. } 
    destruct Hva' as [va' [Eva' Vva]].
    
    eapply pesim_dtsteps.
    + eassumption.
    + econs 2; first by eapply mid_wdet.
      * inst_pairs. econs. by econs. 
      * econs 2; first by eapply mid_wdet.
        { inst_pairs. econs. econs.  eauto. }
        { econs 1. }
    + i; econs. esplits; [| |exact PV| exact FIN]; inst_pairs. eauto. reflexivity.
    + eauto.
    + i; eexists; rrdes RLb. 
      destruct XTR as [Xh]; subst. inv Ec; clean.
      esplits; first 1 last. 
      * left; econs 2; econs 1; inst_pairs; econs 1; econs; eauto.
      * rr; esplits; inst_pairs; s; noemp; eauto.
      * eapply SIM; eauto.
        assert (HK: mm_ksim lw G ren_id (IEcont bx e') (IEcont bx e')).
        { eapply mm_ksim_refl; auto. s. rewrite cons_app. eapply wfexp_strengthen; first exact WFb. 
          s in H. eapply list_mem_false. exact H.
        }
        rrdes HK. 
        destruct (HK g s si i (wfo_nat i) kia kib env' ((inl (inr x), va)::env) bx e'); auto.
        { i. case x0 as [lx1|kx1].
          - destruct (RV (inl lx1)); auto; des. do 2!eexists; eauto.
          - destruct (RV (inr kx1)); auto; des. do 2!eexists; esplits; eauto.
            rewrite list_assoc_cons_neq; auto. injection; i. by clarify.
        }
        des. by eapply SIM0.

  - (*| continuation let before regular let *)

    destruct (LEM (exists ax e', a = IEcont ax e')) as [Hva|Hva]; last first.
    { i. pfold. econs 1. i. destruct XTRa as [Xha], XTRb as [Xhb]; subst.
      econs 1.
      - econs 1. inv CNF; des; subst; rrdes MB; inv Ec.
        eapply IRerr; i; first by auto. inv FH. apply Hva. by eauto. 
      - eauto.
    }
    rrdes Hva; clarify.
    
    destruct (LEM (exists vb, eval ((inr kx, IVclo env (IEcont ax e')) :: env) b = Some vb)) as [Hvb|Hvb]; last first.
    { i. pfold. econs 1. i.
      destruct XTRa as [Xha XTRa], XTRb as [Xhb XTRb]. subst.
      econs.
      - inv CNF; des; subst; rrdes MB; noemp; inv Ec.
        econs 3; econs 1; inst_pairs.
        + econs 1. econs. 
        + econs 2; i; first by auto.
          inv FH. apply Hvb. eauto using EVAL.
      - eauto.
    }
    rrdes Hvb.

    assert (Hvb': exists vb', eval env' b = Some vb'
                           /\ vclos (wlift lw) s (g s ) vb' vb).
    { eapply mm_asim_refl; [| | | |exact Hvb]; eauto.
      i. s in IN; des; clarify; subst. destruct (RV x); auto; des; eauto.
    }
    rrdes Hvb'.

    eapply pesim_dtsteps.
    + eassumption.
    + econs 2; first by eapply mid_wdet.
      * inst_pairs. econs. econs. eassumption.
      * econs 2; first by eapply mid_wdet.
        { inst_pairs. econs. by econs. }
        { econs 1. }
    + i; econs. esplits; [| |exact PV| exact FIN]; inst_pairs. eauto. reflexivity.
    + eauto.
    + i; eexists; rrdes RLb. 
      destruct XTR as [Xh]; subst. inv Ec; clean.
      esplits; first 1 last. 
      * left; econs 2; econs 1; inst_pairs; econs 1; econs; eauto.
      * rr; esplits; inst_pairs; s; noemp; eauto.
      * eapply SIM; eauto.
        assert (HK: mm_ksim lw G ren_id (IEcont ax e') (IEcont ax e')).
        { eapply mm_ksim_refl; auto. }
        rrdes HK. 
        destruct (HK g s si i (wfo_nat i) kia kib ((inl (inr y), vb') :: env') env ax e'); auto.
        { i. case x as [lx1|kx1].
          - destruct (RV (inl lx1)); auto; des. do 2!eexists; eauto.
          - destruct (RV (inr kx1)); auto; des. do 2!eexists; esplits; eauto.
            rewrite list_assoc_cons_neq; auto. injection; i. by clarify.
        }
        des. by eapply SIM0.
    
  - (*| two continuation lets *)

    destruct (LEM (exists ax e', a = IEcont ax e')) as [Hva|Hva]; last first.
    { i. pfold. econs 1. i. destruct XTRa as [Xha], XTRb as [Xhb]; subst.
      econs.
      - econs 1. inv CNF; des; subst; rrdes MB; inv Ec.
        eapply IRerr; i; first by auto. inv FH. apply Hva. by eauto. 
      - eauto.
    }
    rrdes Hva; clarify.

    destruct (LEM (exists bx e', b = IEcont bx e')) as [Hvb|Hvb]; last first.
    { i. pfold. econs 1. i.
      destruct XTRa as [Xha XTRa], XTRb as [Xhb XTRb]. subst.
      econs.
      - inv CNF; des; subst; rrdes MB; noemp; inv Ec.
        econs 3; econs 1; inst_pairs.
        + econs 1. by econs.
        + econs 2; i; first by auto.
          inv FH. apply Hvb. by eauto. 
      - eauto.
    }
    rrdes Hvb; clarify.
    
    eapply pesim_dtsteps.
    + eassumption.
    + econs 2; first by eapply mid_wdet.
      * inst_pairs. econs. by econs.
      * econs 2; first by eapply mid_wdet.
        { inst_pairs. econs. by econs. }
        { econs 1. }
    + i; econs. esplits; [| |exact PV| exact FIN]; inst_pairs. eauto. reflexivity.
    + eauto.
    + i; eexists; rrdes RLb. 
      destruct XTR as [Xh]; subst. inv Ec; clean.
      esplits; first 1 last. 
      * left; econs 2; econs 1; inst_pairs; econs 1; econs; eauto.
      * rr; esplits; inst_pairs; s; noemp; eauto.
      * eapply SIM.
        { assert (HK: mm_ksim lw G ren_id (IEcont ax e') (IEcont ax e')).
          { eapply mm_ksim_refl; by auto. }
          rrdes HK. 
          destruct (HK g s si i (wfo_nat i) kia kib ((inr ky, IVclo env' (IEcont bx e'0)) :: env') env ax e'); auto.
          { i. rewrite list_assoc_cons_neq; first by eauto. rr; i; by clarify. }
          des. by eapply SIM0.
        }
        { assert (HK: mm_ksim lw G ren_id (IEcont bx e'0) (IEcont bx e'0)).
          { eapply mm_ksim_refl; auto. s. rewrite cons_app. eapply wfexp_strengthen; first exact WFb. 
            s in H. eapply list_mem_false. exact H.
          }
          rrdes HK. 
          destruct (HK g s si i (wfo_nat i) kia kib env' ((inr kx, IVclo env (IEcont ax e')) :: env) bx e'0); auto.
          { i. rewrite list_assoc_cons_neq; first by eauto. rr; i; by clarify. }
          des. by eapply SIM0.
        }
Qed.



(** * Analysis. *)


(** ** Annotation type. *)

Definition commute1_ann: annT.
  refine {| _annT := unit
          ; perform u := commute
         |}.
  - ii; eauto using commute_wfexp.
  - ii; econs; eauto using commute_psim.
Defined.

Definition commute_ann: annT :=
  annT_list commute1_ann.


(** ** Algorithm. *)

Module otf := opt_transform.

Section algorithm.

(** Order on expressions.
    For the algorithm to be useful, le_exp should satisfy the following:
    le_exp a b = true /\ le_exp b a = true -> a = b.
*)
Variable le_exp: exp -> exp -> bool.

Fixpoint analyze_ (e: exp) : aexp commute_ann * aexp exp :=
  match e with
    | IElet ea xa e => 
      let aneb := analyze_ e in
      let anea := analyze_ ea in
      let (annot, exp_and_var) := 
          match aneb with
              | (otf.IElet _ eb xb e, otf.IElet real_eb _ _ _)  =>
                if le_exp ea real_eb then
                  ([], (ea, xa)) 
                else
                  ([()], (real_eb, xb))
              | _ => ([], (ea, xa))
          end
      in 
      let (new_top_exp, new_top_var) := exp_and_var
      in (otf.IElet annot (fst anea) xa (fst aneb),
          otf.IElet new_top_exp (if annot then snd aneb else (snd anea))
                    new_top_var (if annot then (snd anea) else snd aneb))
    | IEnew x bx e0 => 
      let e0' := analyze_ e0 in
      (otf.IEnew x bx (fst e0'), otf.IEnew x bx (snd e0'))
    | IEread x bx e0 =>
      let e0' := analyze_ e0 in
      (otf.IEread x bx (fst e0'), otf.IEread x bx (snd e0'))
    | IEwrite x y e0 => 
      let e0' := analyze_ e0 in
      (otf.IEwrite x y (fst e0'), otf.IEwrite x y (snd e0'))
    | IEinput bx e0 => 
      let e0' := analyze_ e0 in
      (otf.IEinput bx (fst e0'), otf.IEinput bx (snd e0'))
    | IEoutput x e0 => 
      let e0' := analyze_ e0 in
      (otf.IEoutput x (fst e0'), otf.IEoutput x (snd e0'))
    | IEifnz x ea eb => 
      let ea' := analyze_ ea in
      let eb' := analyze_ eb in
      (otf.IEifnz x (fst ea') (fst eb'), otf.IEifnz x (snd ea') (snd eb'))
    | IEcase x bx ea eb => 
      let ea' := analyze_ ea in
      let eb' := analyze_ eb in
      (otf.IEcase x bx (fst ea') (fst eb'), otf.IEcase x bx (snd ea') (snd eb'))
    | IEret xk x => 
      (otf.IEret _ xk x, otf.IEret _ xk x)
    | IEapp xf x xk => 
      (otf.IEapp nil xf x xk, otf.IEapp IEunit xf x xk) 
    | IEinst xf xk => 
      (otf.IEinst nil xf xk, otf.IEinst IEunit xf xk) 
    | IEpair x y => 
      (otf.IEpair _ x y, otf.IEpair _ x y)
    | IEinl x =>
      (otf.IEinl _ x, otf.IEinl _ x)
    | IEinr x =>
      (otf.IEinr _ x, otf.IEinr _ x)
    | IEbop bop x y => 
      (otf.IEbop _ bop x y, otf.IEbop _ bop x y)
    | IEfst x =>
      (otf.IEfst _ x, otf.IEfst _ x)
    | IEsnd x =>
      (otf.IEsnd _ x, otf.IEsnd _ x)
    | IEfix bf bx bk e0 =>
      let e0' := analyze_ e0 in
      (otf.IEfix bf bx bk (fst e0'), otf.IEfix bf bx bk (snd e0'))
    | IEgen bk e0 => 
      let e0' := analyze_ e0 in
      (otf.IEgen bk (fst e0'), otf.IEgen bk (snd e0'))
    | IEcont bx e0 => 
      let e0' := analyze_ e0 in
      (otf.IEcont bx (fst e0'), otf.IEcont bx (snd e0'))
    | IEunit =>
      (otf.IEunit _, otf.IEunit _)
    | IEnat n =>
      (otf.IEnat _ n, otf.IEnat _ n)
    | IErefeq x y =>
      (otf.IErefeq _ x y, otf.IErefeq _ x y)
  end.

Definition analyze e :=
  fst (analyze_ e).


Theorem analyze_erase: forall e,
  aerase (analyze e) = e.
Proof.
  unfold analyze.
  induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  destruct e;
  try (simpl in *; by auto);
  try (simpl in *; rewrite (IH e); try omega; try reflexivity; by rewrite H0 H1);
  try (simpl in *; 
        rewrite (IH e1); try omega; try reflexivity; 
        rewrite (IH e2); try omega; try reflexivity; 
         by rewrite H0 H1 H2 H3).
  hexploit (IH e1); [s; omega|].
  hexploit (IH e2); [s; omega|].
  intros E2 E1; s.
  destruct (analyze_ e2); destruct a; simpl in *; try rewrite E1 E2; auto.
  destruct a0; ss; rewrite ?E1 ?E2; auto.
  destruct (le_exp _ _); s; by rewrite E1 E2.
Qed.

End algorithm.


(** * Module transformation. *)

Definition commute_mod (le: exp -> exp -> bool) : modtf :=
  fun _ => transform_mod (analyze le).

Theorem commute_wfmod le: forall FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m),
  mid_wfmod FI (commute_mod le FI m) FO /\
  uniqvar (commute_mod le FI m).
Proof.
  i; eapply transform_wfmod; eauto using analyze_erase.
Qed.

Theorem commute_msim le: forall FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m),
  rtc (msim mm_gw FI FO) (commute_mod le FI m) m.
Proof.
  i; eapply transform_msim; eauto using analyze_erase.
Qed.  

Theorem commute_correct le:
  <<WF: modtf_wf (commute_mod le)>> /\
  <<SIM: modtf_sim (commute_mod le)>>.
Proof.
  split; ii; eauto using commute_wfmod, commute_msim.
Qed.
