(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

(** Simple dead code elimination. *)

Require Import Arith Bool List Program.
Require Import countable vgtac basis lang_common lang_spec lang_mid.
Require Import model_common model_untyped model_properties_untyped gw_untyped mm.
Require Import ssr pilsner_common.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



(** * Algorithm.

    The only interesting cases in [dce] are [IElet] and [IEread]: if
    the variable being bound is never used (syntactically), the
    operation is simply replaced by its body.

    For efficiency, we don't use a separate function for collecting
    the used variables but do it on the fly (second component of result).
*)

Fixpoint dce (e: exp) : exp * list var :=
  match e with
    | IElet ea bx eb
      => let (eb', xsb) := dce eb in
         if list_mem eqdec (tkvar2var bx) xsb then
           
           let (ea', xsa) := dce ea in
           (IElet ea' bx eb', xsa ++ xsb)
         else
           
           (eb', xsb)

    | IEread x bx ea
      => let (ea', xsa) := dce ea in
         if list_mem eqdec (inl (inr bx)) xsa then
           
           (IEread x bx ea', [inl x] ++ xsa)
         else
           
           (ea', xsa)

    | IEnew x bx ea
      => let (ea', xsa) := dce ea in
         if list_mem eqdec (inl (inr bx)) xsa then
           
           (IEnew x bx ea', [inl x] ++ xsa)
         else
           
           (ea', xsa)

    | IEpair x y
      => (e, [inl x; inl y])

    | IEret k x
    | IEinst x k
      => (e, [inl x; inr k])

    | IEinl x
    | IEinr x
    | IEfst x
    | IEsnd x
      => (e, [inl x])

    | IEbop _ xn yn
      => (e, map inl (inl_sng xn) ++ map inl (inl_sng yn))

    | IEapp x y k
      => (e, [inl x; inl y; inr k])

    | IEoutput x ea
      => let (ea', xsa) := dce ea in
         (IEoutput x ea', [inl x] ++ xsa)

    | IEwrite x y ea
      => let (ea', xsa) := dce ea in
         (IEwrite x y ea', [inl x; inl y] ++ xsa)

    | IEgen bx ea
      => let (ea', xsa) := dce ea in
         (IEgen bx ea', xsa)
    | IEcont bx ea
      => let (ea', xsa) := dce ea in
         (IEcont bx ea', xsa)
    | IEinput bx ea
      => let (ea', xsa) := dce ea in
         (IEinput bx ea', xsa)

    | IEfix bf bx bk ea
      => let (ea', xsa) := dce ea in
         (IEfix bf bx bk ea', xsa)

    | IEifnz x ea eb
      => let (ea', xsa) := dce ea in
         let (eb', xsb) := dce eb in
         (IEifnz x ea' eb', [inl x] ++ xsa ++ xsb)
    | IEcase x bx ea eb
      => let (ea', xsa) := dce ea in
         let (eb', xsb) := dce eb in
         (IEcase x bx ea' eb', [inl x] ++ xsa ++ xsb)
    | IErefeq xn yn
      => (e, [inl xn; inl yn])

    | _
      => (e, [])
  end.


Definition dce_mod: modtf :=
  fun _ => modmap (fst<*>dce).



(** * Properties. *)

Lemma dce_vars: forall e x
    (IN: list_mem eqdec x (snd (dce e)) = false),
  ~ In x (vars (fst (dce e))).
Proof.
  induction e; ss; i
  ; try (by try destruct (dce _) eqn:E; ss; repeat (des_ifH IN; vauto)
         ; des; vauto; eauto using list_mem_false).
  - eapply list_mem_false; eauto.
  - destruct (dce e) eqn:E; des_ifH IN; [|by eauto].
    ss; des; des_ifH IN; vauto; []; eauto.
  - destruct (dce e) eqn:E; des_ifH IN; [|by eauto].
    ss; des; des_ifH IN; vauto; []; eauto.
  - destruct (dce e1) eqn:E1; destruct (dce e2) eqn:E2.
    ss; repeat (des_ifH IN; vauto); des; vauto.
    rewrite list_mem_app in IN; apply orb_false_iff in IN; des.
    apply in_app_or in FH; des; eauto using list_mem_false.
  - destruct (dce e1) eqn:E1; destruct (dce e2) eqn:E2.
    ss; repeat (des_ifH IN; vauto); des; vauto. 
    rewrite list_mem_app in IN; apply orb_false_iff in IN; des.
    apply in_app_or in FH; des; eauto using list_mem_false.
  - destruct (dce e1) eqn:E1; destruct (dce e2) eqn:E2.
    des_ifH IN; ss.
    + rewrite list_mem_app in IN; apply orb_false_iff in IN; des.
      apply in_app_or in FH; des; eauto using list_mem_false.
    + eauto using list_mem_false.
Qed.

(** [sublist]s are used as a tool to compose [uniq]ueness. *)

Lemma wfatom_wf: forall e G
    (WF: wfatom G e)
    (IH: forall e', size e' < size e -> forall G, wfexp G e' ->
           wfexp G (fst (dce e')) /\ sublist (bvars (fst (dce e'))) (bvars e')),
  wfatom G (fst (dce e)) /\
  sublist (bvars (fst (dce e))) (bvars e).
Proof.
  destruct e; vauto; i; ss; rrdes WF; vauto
  ; (by destruct (dce e) eqn:E; ss
     ; exploit (IH e); eauto; inst
     ; rewrite E; i; des; split; eauto using sublist_app
     ; econs 2; eauto).
Qed.

Lemma dce_wf: forall e G
    (WF: wfexp G e),
  wfexp G (fst (dce e)) /\
  sublist (bvars (fst (dce e))) (bvars e).
Proof.
  induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  destruct e; vauto; i; ss; des
  ; try (by destruct (dce _) eqn:E; ss
         ; hexploit (IH e); eauto; [omega|]
         ; inst; rewrite E; i; des; eauto).
  - assert (FV:=dce_vars e (inl (inr bx))).
    destruct (dce e) eqn:E; ss.
    hexploit (IH e); eauto.
    rewrite E; i; des; eauto; unfold var in FV.
    des_if; ss; splits; eauto.
    rewrite <-(app_nil_l G).
    eapply wfexp_strengthen; eauto.
  - assert (FV:=dce_vars e (inl (inr bx))).
    destruct (dce e) eqn:E; ss.
    hexploit (IH e); eauto.
    rewrite E; i; des; eauto; unfold var in FV.
    des_if; ss; splits; eauto.
    rewrite <-(app_nil_l G).
    eapply wfexp_strengthen; eauto.
  - destruct (dce _) eqn:E; ss.
    hexploit (IH e); eauto.
    rewrite E; i; des; eauto.
  - destruct (dce _) eqn:E; ss.
    hexploit (IH e); eauto.
    rewrite E; i; des; eauto.
  - destruct (dce _) eqn:E; ss.
    hexploit (IH e); eauto.
    rewrite E; i; des; eauto.
  - destruct (dce e1) eqn:E1; ss.
    destruct (dce e2) eqn:E2; ss.
    hexploit (IH e1); eauto; [omega|].
    hexploit (IH e2); eauto; [omega|].
    rewrite E1 E2; i; des; eauto using sublist_app.
  - destruct (dce e1) eqn:E1; ss.
    destruct (dce e2) eqn:E2; ss.
    hexploit (IH e1); eauto; [omega|].
    hexploit (IH e2); eauto; [omega|].
    rewrite E1 E2; i; des; setoid_rewrite cons_app at 3 4
    ; eauto 10 using sublist_app.
  - assert (FV:=dce_vars e2 (tkvar2var bx)).
    destruct (dce e2) eqn:E2; ss; des_if.
    + destruct (dce e1) eqn:E1; ss.
      destruct bx; des
      ; (hexploit wfatom_wf; eauto
         ; [by i; eapply (IH e'); eauto; omega|]
         ; hexploit (IH e2); eauto; [omega|]
         ; fold (fst (e0, l0)); rewrite <-E1
         ; rewrite E1 E2; i; des; eauto using sublist_app).
    + destruct bx; des
      ; (rewrite cons_app;
         hexploit (IH e2); eauto; [omega|]
         ; rewrite E2; i; des; split
         ; try (by change (bvars (fst (e, l))) with ([] ++ [] ++ bvars e)
                ; auto using sublist_app)
         ; rewrite <-(app_nil_l G)
         ; s in H0; try rewrite <-(app_nil_l (_ :: _)) in H0
         ; eapply wfexp_strengthen; ss; eauto).
Qed.

Lemma dce_tl_wf: forall G e
    (WF: wfatom G e)
    (GT: is_good_thunk e)
    (UV: uniq (bvars e)),
  wfatom G (fst (dce e)) /\
  is_good_thunk (fst (dce e)) /\
  sublist (bvars (fst (dce e))) (bvars e).
Proof.
  i; destruct e; vauto
  ; by (s in WF; s; destruct (dce _) eqn:E; s
        ; hexploit dce_wf; [by eauto|]; inst
        ; rewrite E; i; des; ss; splits; eauto
        
        ; s; destruct (dce _); ii; inv FH; ss; des; apply UV; auto).
Qed.

(** Dead code elimination preserves well-formedness as well as the
    unique variable condition.
*)
Theorem dce_wfmod: forall FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m),
  mid_wfmod FI (dce_mod FI m) FO /\
  uniqvar (dce_mod FI m).
Proof.
  i; apply modmap_wfmod; eauto; []; i.
  dup UN; apply uniq_appD in UN; des.
  hexploit dce_tl_wf; eauto; []; i; des.
  splits; auto.
  eapply uniq_sublist; eauto using sublist_app.
Qed.



(** * Correctness. *)


(** ** The local world used for the proof. *)

Program Definition mylwsts: transys :=
  {| state := heap  
   ; ext    s' s := True
   ; extpub s' s := True
  |}.

Program Definition mylwconf: method_conf mid_lsi mid_lsi (tsprod (gwsts mm_gw) mylwsts) :=
  {| confrel g s ca cb := ca = cemp /\ cb = (snd s, Nothing) |}.

Program Definition mylw: world_loca mm_gw :=
  {| lwwfo := gwfo
   ; lwsts := mylwsts; lwconf := mylwconf |}.
Next Obligation.
  unfold mid_extra; unnw; subst; ss; unfold mid_cemp; eauto.
Qed.

Lemma mylw_stable: stable mylw.
Proof.
  ii; rrdes CR; rr; esplits; s; eauto.
Qed.


(** ** Key properties for the non-trivial cases of the transformation. *)

Lemma dce_esim_let: forall G e1 bx e2
    (IH: mm_psim' mylw (list_minus eqdec G [inl (inr bx)])
                 (fst (dce e2)) e2),
  mm_psim' mylw G (fst (dce e2)) (IElet e1 (inl bx) e2).
Proof.
  i; rrdes IH; rr; splits; vauto; []; eexists; i; s.
  eapply pesim_acquire; i; pfold; econs; i.
  invc CNF XTRa XTRb; des; subst
  ; dup HR; destruct HR as (? & ? & ? & ? & WHX & WHY & ? & ?)
  ; rrdes WHY; apply mm_confrel_shape in WHX; des
  ; destruct MA as (? & ? & Eca & Ema & PVa & FINa)
  ; destruct MB as (? & ? & Ecb & Emb & PVb & FINb)
  ; subst; s in Eca; s in Ecb; noemp; unnw; clarify.
  destruct (LEM (forall evt hb'' envb' e',
                   ~ reduce_ok evt ((hb' ## snd s) ## x0) envb (IElet e1 (inl bx) e2)
                                   hb'' envb' e')).
  { eapply esim_err; [by econs; eapply IRerr; eauto|].
    rr; i; inv FH; des; clarify; rrdes PV; vauto. }
  
  repeat (eapply not_all_ex_not in f; des); clear f; inv n3.
  hexploit pesim_stepb_obsolete; last first.
  - intro SIM; punfold SIM; invc SIM.
    eapply ESIM; rr; eauto; econs; eauto
    ; inst_pairs; s; noemp; rr; esplits; eauto.
  - rewrite mid_csplit; instantiate (1:=_#_).
    eapply pesim_release; eauto; inst_pairs.
    + eapply (IH1 g s si enva ((inl (inr bx), v) :: envb)); eauto.
      * i; apply in_list_minus in IN; des.
        s in IN0; des_ifH IN0; vauto; clear IN0.
        rewrite list_assoc_cons_neq; auto.
      * i; apply in_list_minus in IN; des; clear IN0.
        rewrite list_assoc_cons_neq; auto; vauto.
    + rr; eauto.
    + rr; eauto.
  - i; invc XTR; s in RL; rrdes RL; clarify; esplits.
    + econs 1; inst_pairs; eapply IRok; eauto using IRlet.
    + s; noemp; rr; eauto 10.
Grab Existential Variables.
  exact 0.
Qed.

Lemma dce_esim_letcont: forall G e1 bx e2
    (IH: mm_psim' mylw (list_minus eqdec G [inr bx])
                 (fst (dce e2)) e2),
  mm_psim' mylw G (fst (dce e2)) (IElet e1 (inr bx) e2).
Proof.
  i; rrdes IH; rr; splits; vauto; []; eexists; i.
  eapply pesim_acquire; i; pfold; econs; i.
  invc CNF XTRa XTRb; des; subst
  ; dup HR; destruct HR as (? & ? & ? & ? & WHX & WHY & ? & ?)
  ; rrdes WHY; apply mm_confrel_shape in WHX; des
  ; destruct MA as (? & ? & Eca & Ema & PVa & FINa)
  ; destruct MB as (? & ? & Ecb & Emb & PVb & FINb)
  ; subst; s in Eca; s in Ecb; noemp; unnw; clarify.
  destruct (LEM (forall evt hb'' envb' e',
                   ~ reduce_ok evt ((hb' ## snd s) ## x0) envb (IElet e1 (inr bx) e2)
                                   hb'' envb' e')).
  { eapply esim_err; [by econs; eapply IRerr; eauto|].
    rr; i; inv FH; des; clarify; rrdes PV; vauto. }
  
  repeat (eapply not_all_ex_not in f; des); clear f; inv n3.
  hexploit pesim_stepb_obsolete; last first.
  - intro SIM; punfold SIM; invc SIM.
    eapply ESIM; rr; eauto; econs; eauto
    ; inst_pairs; s; noemp; rr; esplits; eauto.
  - rewrite mid_csplit; instantiate (1:=_#_).
    eapply pesim_release; eauto; inst_pairs.
    + eapply (IH1 g s si enva ((inr bx, IVclo envb (IEcont bx0 e')) :: envb)); eauto.
      * i; apply in_list_minus in IN; des; clear IN0.
        rewrite list_assoc_cons_neq; auto; vauto.
      * i; apply in_list_minus in IN; des.
        s in IN0; des_ifH IN0; vauto; clear IN0.
        rewrite list_assoc_cons_neq; auto.
    + rr; eauto.
    + rr; eauto.
  - i; invc XTR; s in RL; rrdes RL; clarify; esplits.
    + econs 1; inst_pairs; eapply IRok; eauto using IRlet.
    + s; noemp; rr; eauto 10.
Grab Existential Variables.
  exact 0.
Qed.

Lemma stepb_low: forall La Lb (w: world La Lb) mb mb' s s'
    es i kia kib g si cfa ma cfb
    (ST: rtc (step Lb etau) mb mb')
    (EXT: ext _ s' s)
    (SIM: esim_main (w:=w)
             (esim_progress es i kia kib g si \7/ esim_call es kia kib g si)
             kia kib g si s' cfa ma cfb mb'),
  esim_main
    (esim_progress es i kia kib g si \7/ esim_call es kia kib g si)
    kia kib g si s cfa ma cfb mb.
Proof.
  i; eauto using esim_main_mon, esim_main_taus.
Qed.                            

Lemma dce_esim_new: forall G e x bx
    (IH: mm_psim' mylw (list_minus eqdec G [inl (inr bx)])
                 (fst (dce e)) e),
  mm_psim' mylw G (fst (dce e)) (IEnew x bx e).
Proof.
  i; rrdes IH; rr; splits; vauto; []; eexists; i.
  eapply pesim_acquire; i; pfold; econs; i.
  inv CNF XTRa XTRb; des; subst
  ; dup HR; destruct HR as (? & ? & ? & ? & WHX & WHY & ? & ?)
  ; rrdes WHY; apply mm_confrel_shape in WHX; des
  ; destruct MA as (? & ? & Eca & Ema & PVa & FINa)
  ; destruct MB as (? & ? & Ecb & Emb & PVb & FINb)
  ; subst; s in Eca; s in Ecb; noemp; unnw; clarify.
  destruct (LEM (forall evt hb'' envb' e',
                   ~ reduce_ok evt ((hb' ## snd s) ## x1) envb (IEnew x bx e)
                                   hb'' envb' e')).
  { eapply esim_err; [by econs; eapply IRerr; eauto|].
    rr; i; inv FH; des; clarify; rrdes PV; vauto. }
  repeat (eapply not_all_ex_not in f; des); clear f; inv n3.
  hexploit (@pmfinite_alloc _ _ _ ((hb' ## snd s) ## x1)); eauto; []; i; des.
  set (s' := (fst s, {x2 |-> v} ## snd s) : state (wlift mylw)).
  assert (EXT': extpub _ s' s) by (split; auto).
  eapply stepb_low.
  { econs 1; inst_pairs; eapply IRok, IRnew; eauto. }
  { eauto. }
  hexploit (IH1 g s' si enva ((inl (inr bx), IVloc x2) :: envb)); eauto.
  { i; apply in_list_minus in IN; des.
    rewrite list_assoc_cons_neq.
    - apply RV in IN; des; esplits; eauto using vclos_mon, (gk_mon GK).
    - ii; ss; des_ifs; vauto.
  }
  { i; apply in_list_minus in IN; des.
    rewrite list_assoc_cons_neq.
    - apply RK in IN; des; esplits; eauto.
      eapply ksim_mon; eauto.
    - ii; ss; des_ifs; vauto.
  }
  i; punfold H; invc H.
  exploit (ESIM GK0 _ _ XTRa XTRb); [|by eauto].
  invc CNF; econs; try (split; s); eauto; last first.
  - inst_pairs; instantiate (1 := Nothing).
    rrdes MB; s in Ec; clarify; noemp.
    rr; esplits; eauto; s; noemp; eauto.
    rewrite pmplus_assoc; eauto.
  - inst_pairs; instantiate (1 := Nothing).
    ss; noemp; eauto.
  - rrdes HR0.
    rr; esplits.
    + eapply confrel_mon; eauto.
      unfold s' in *; rrdes EXT'.
      s; i; eapply (gk_mon GK); ss; eauto.
    + rr; eauto.
    + rrdes WHY; subst; noemp; eauto.
    + destruct hbx; unfold s'; rrdes WHY; ss; subst; noemp; clarify.
      rewrite -H1 pmplus_swap; eauto.
Qed.


Lemma dce_esim_read: forall G e x bx
    (IH: mm_psim' mylw (list_minus eqdec G [inl (inr bx)])
                 (fst (dce e)) e),
  mm_psim' mylw G (fst (dce e)) (IEread x bx e).
Proof.
  i; rrdes IH; rr; splits; vauto; []; eexists; i.
  eapply pesim_acquire; i; eapply pesim_stepb.
  i; invc XTR; rrdes HR RL.
  hexploit lwconf_extra; eauto; []; i; des; invc X X0.
  hexploit mm_confrel_shape; eauto; []; i; des.
  clarify; s in Ec; noemp; clarify.
  destruct (LEM (forall evt hb'' envb' e',
                   ~ reduce_ok evt ((hb' ## x2) ## x0) envb (IEread x bx e)
                     hb'' envb' e')).
  { esplits; [econs; eapply IRerr; eauto|]; left; eauto. }
  repeat (eapply not_all_ex_not in f; des); clear f; inv n3.
  esplits; [|right; esplits].
  { econs 1; inst_pairs; eapply IRok, IRread; eauto. }
  { inst_pairs; s; noemp; rr; esplits; s; eauto. }
  s; noemp; rewrite 2!mid_csplit.
  apply pesim_release; try (by rr; eauto); [|by rr; esplits; eauto].
  eapply IH1; eauto.
  - i; apply in_list_minus in IN; des.
    rewrite list_assoc_cons_neq; eauto; vauto.
    ii; ss; des_ifs; vauto.
  - i; apply in_list_minus in IN; des.
    rewrite list_assoc_cons_neq; eauto.
    i; vauto.
Grab Existential Variables.
  exact cemp. exact 0.
Qed.


(** ** Putting things together. *)

Lemma dce_correct_asim: forall G e
    (WF: wfatom G (fst (dce e)))
    (IH: forall e', size e' < size e -> forall G,
           wfexp G (fst (dce e')) -> mm_psim' mylw G (fst (dce e')) e'),
  mm_asim' mylw G (fst (dce e)) e.
Proof.
  i; destruct e; vauto; ss; des; auto with mm_compat';
  specialize (IH e); desf; ss; eauto with mm_compat'.
Qed.

Lemma dce_correct_ksim: forall G e
    (WF: wfatom G (fst (dce e)))
    (IH: forall e', size e' < size e -> forall G,
           wfexp G (fst (dce e')) -> mm_psim' mylw G (fst (dce e')) e'),
  mm_ksim' mylw G (fst (dce e)) e.
Proof.
  ii; destruct e; vauto; ss.
  destruct (dce eb') eqn:E; ss.
  hexploit (IH eb'); try rewrite E; eauto; i.
  esplits; [by auto|].
  hexploit mm_compat_cont'; eauto; intro SIM.
  unfold mm_ksim' in SIM.
  hexploit SIM; last first; eauto; i; des; eauto.
Qed.

Lemma dce_correct_psim: forall e G
    (WF: wfexp G (fst (dce e))),
  mm_psim' mylw G (fst (dce e)) e.
Proof.
  intro; induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  destruct e; vauto; i; ss; subst
  ; try (destruct (dce e) eqn:E)
  ; try (destruct (dce e1) eqn:E1, (dce e2) eqn:E2)
  ; ss; vauto; ss; des
  ; try (hexploit (IH e); try rewrite E; eauto; i; [])
  ; try (hexploit (IH e1); try rewrite E1; eauto; i; [omega|])
  ; try (hexploit (IH e2); try rewrite E2; eauto; i; [omega|])
  ; auto with mm_compat'.
  - des_ifs.
    + invc WF; hexploit (IH e); try rewrite E
      ; eauto using mm_compat_new', mylw_stable.
    + rewrite -E; apply dce_esim_new.
      eapply (IH e); eauto.
      rewrite E; eapply wfexp_strengthen'; eauto.
      rewrite -E; eapply dce_vars; rewrite E; eauto.
  - des_ifs.
    + invc WF; hexploit (IH e); try rewrite E
      ; eauto using mm_compat_read'.
    + rewrite -E; apply dce_esim_read.
      eapply (IH e); eauto.
      rewrite E; eapply wfexp_strengthen'; eauto.
      rewrite -E; eapply dce_vars; rewrite E; eauto.
  - des_ifs.
    + s; destruct bx; s in WF; des
      ; fold (fst (e, l)); rewrite <-E1
      ; fold (fst (e, l)) in WF; rewrite <-E1 in WF.
      * hexploit (IH e2); try rewrite E2; eauto; [omega|i].
        apply mm_compat_let'; eauto.
        apply dce_correct_asim; auto; i.
        apply (IH e'); eauto; omega.
      * hexploit (IH e2); try rewrite E2; eauto; [omega|i].
        apply mm_compat_letcont'; eauto.
        apply dce_correct_ksim; auto; i.
        apply (IH e'); eauto; omega.
    + fold (snd (e0, l0)) in Heq; rewrite <-E2 in Heq.
      eapply wfexp_strengthen' in WF; [|by rewrite <-E2; apply dce_vars; eauto].
      hexploit (IH e2); try rewrite E2; eauto; [omega|i].
      rewrite <-E2; destruct bx.
      * eapply dce_esim_let; eauto; rewrite E2; eauto.
      * eapply dce_esim_letcont; eauto; rewrite E2; eauto.
Qed.

Theorem dce_correct_msim: forall FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m),
  rtc (msim mm_gw FI FO) (dce_mod FI m) m.
Proof.
  i; econs 1; ginduction WF; eauto using mm_msim_nil.
  i; apply uniq_appD in UV; fold bvars_mod in UV; des.
  hexploit IHWF; eauto; []; i.
  eapply mm_msim_cons; eauto using mylw_stable.
  - destruct e; vauto.
    + s; destruct (dce _) eqn:E; econs 1; change e0 with (fst (e0, l)); rewrite -E.
      s in WFE; rewrite -map_map purge_map_inl map_map in WFE.
      eapply mm_psim_old; eapply dce_correct_psim; eauto; eapply dce_wf; eauto.
    + s; destruct (dce _) eqn:E; econs 2; change e0 with (fst (e0, l)); rewrite -E.
      s in WFE; rewrite -map_map purge_map_inl map_map in WFE.
      eapply mm_psim_old; eapply dce_correct_psim; eauto; eapply dce_wf; eauto.
  - inst_pairs; i; rr; s; unfold mid_cemp; eauto.
Qed.


Theorem dce_correct:
  <<WF: modtf_wf dce_mod>> /\
  <<SIM: modtf_sim dce_mod>>.
Proof.
  split; ii; eauto using dce_wfmod, dce_correct_msim.
Qed.
