(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

(** Countability of types and sets *)

Require Import Compare_dec Le Image List Ensembles Mult Program.
Require Import hpattern vgtac myclassical.

Set Implicit Arguments.
Unset Strict Implicit.
Open Scope string_scope.



(** * Preliminary definitions and properties. *)


Definition surjective A B (f: A -> B) : Prop := forall b,
  exists a, b = f a.

Definition ndisj A (N1 N2: A -> Prop) := forall n,
  N1 n -> N2 n -> False.

Definition nunion A (N1 N2: A -> Prop) :=
  fun n => N1 n \/ N2 n. 

Fixpoint exponential (n m: nat) : nat :=
  match m with
    | 0 => 1
    | S m' => exponential n m' * n
  end.

Fixpoint nat_pair (nm: nat*nat) : nat :=
  exponential 2 (fst nm) * exponential 3 (snd nm).


Lemma ndisj_mon : forall A (N1' N2' N1 N2: A -> Prop)
    (DIS: ndisj N1' N2')
    (LT1: forall a (EL: N1 a), N1' a)
    (LT2: forall a (EL: N2 a), N2' a),
  ndisj N1 N2.
Proof.
  red; intros; eauto.
Qed.

Lemma even_or_odd: forall n,
  exists i, n = 2 * i \/ n = S (2 * i).
Proof.
  by intros; induction n; des; subst; eauto; exists (S i); omega.
Qed.

Lemma exponential2_inv: forall n m,
  exponential 2 n = m * 3 -> False.
Proof.
  induction n; intros; simpls; [by omega|].
  by destruct (even_or_odd m); des; subst; [specialize (IHn x)|]; omega.
Qed.

Lemma exponential2_inj:
  injective _ _ (exponential 2).
Proof.
  intro n; induction n; intros []; intros; simpls; try omega.
  by f_equal; eapply IHn; omega.
Qed.

Lemma nat_pair_inj:
  injective _ _ nat_pair.
Proof.
  intros (n & m); revert n; induction m; i; destruct y; ss.
  - destruct n1; ss.
    + by rewrite !mult_1_r in H; eapply exponential2_inj in H; rewrite H.
    + by exfalso; rewrite mult_1_r, mult_assoc in H
      ; eauto using exponential2_inv.
  - destruct n1; ss.
    + by exfalso; rewrite mult_1_r, mult_assoc in H
      ; eauto using exponential2_inv.
    + exploit (IHm n (n0,n1)); [|by intro X; inv X].
      ss; rewrite !mult_assoc in H; omega.
Qed.



(** * Countably infinite types. *)


(** ** Two alternative definitions. *)

Definition countably_infinite (A: Type) :=
  exists f: A -> nat, exists g: nat -> A,
    (forall a, g (f a) = a) /\
    (forall n, f (g n) = n).

Definition countably_infinite2 (A: Type) :=
  exists f: A -> nat, exists g: nat -> A,
    injective _ _ f /\ injective _ _ g.


(** ** Proof that they are equivalent. *)

Lemma inj_srj_impl_bij: forall A B (f: A -> B)
    (INJ: injective _ _ f)
    (SRJ: surjective f),
  exists g, (forall a, g (f a) = a) /\ (forall b, f (g b) = b).
Proof.
  intros; eapply functional_choice in SRJ; des; eauto.
Qed.

Fixpoint count_prop (P: nat -> Prop) (n: nat) : nat :=
  match n with
    | 0 => 0
    | S n' => match LEM (P n') with
                | left _ => 1 |
                right _ => 0
              end + count_prop P n'
  end.

Definition compactify A (f: A -> nat) (a : A) : nat :=
  count_prop (fun n => exists a, n = f a) (f a).

Lemma count_prop_mon: forall P n n'
    (LE: n <= n'),
  count_prop P n <= count_prop P n'. 
Proof.
  intros; induction LE; simpl; omega.
Qed.

Lemma compactify_le: forall A (f: A -> nat) a n
    (LE: n <= compactify f a),
  exists a', n = compactify f a'.
Proof.
  unfold compactify; i; eapply le_lt_eq_dec in LE.
  destruct LE as [LE|LE]; eauto; revert LE.
  generalize (f a); clear a; intros m LE.
  induction m; intros; simpls; [by exfalso; omega|].
  destruct LEM; eauto; eapply le_lt_eq_dec in LE; destruct LE.
  - by eapply IHm; omega.
  - by des; subst; exists a; omega.
Qed.

Lemma nat_map_inj_impl_unbounded: forall (f: nat -> nat) m
    (INJ: injective _ _ f),
  exists n, f n >= m.
Proof.
  i; eapply NNPP; intro X; assert (Y := not_ex_all_not _ _ X); clear X; ss.

  assert (CARDSEG: forall m, cardinal _ (fun n => n < m) m).
  {
    clear; induction m.
    - setoid_rewrite (@Extensionality_Ensembles nat); [by econs|].
      split; ii; unfold In in *; [exfalso; omega|destruct H].
    - setoid_rewrite (@Extensionality_Ensembles nat).
      + by econs; eauto; unfold In; instantiate (1:= m); omega.
      + split; ii; unfold In in *.
        * assert (X: x = m \/ x < m) by omega.
          by des; subst; [econstructor 2| econstructor 1].
        * by destruct H; red in H; [omega|inv H].
  }

  assert (CARDIM := @injective_preserves_cardinal
                      nat nat (fun n => n < S m) _ _ INJ (CARDSEG (S m))).

  assert (INCL: Included nat (Im _ _ (fun n => n < S m) f) (fun n => n < m)).
  {
    repeat intro; inv H; unfold In in *; specialize (Y x0); omega.
  }

  assert (CARD: exists n, cardinal nat (Im nat nat (fun n => n < S m) f) n).
  {
    eauto using finite_cardinal, cardinal_finite, Finite_downward_closed.
  }

  des; specialize (CARDIM _ CARD); subst.
  assert (S m <= m) by eauto using incl_card_le.
  omega.
Qed.

Lemma compactify_inj: forall A (f: A -> nat)
    (INJ: injective _ _ f),
  injective _ _ (compactify f).
Proof.
  red; intros; eapply INJ, NNPP; red; intros; unfold compactify in *.
  destruct (lt_eq_lt_dec (f x) (f y)) as [[CMP|CMP]|CMP]; try contradiction.
  - assert (LE: S (f x) <= f y) by omega.
    eapply (count_prop_mon (fun n => exists a, n = f a)) in LE.
    rewrite <-H in LE; simpl in LE; destruct LEM; [omega|eauto].
  - assert (LE: S (f y) <= f x) by omega.
    eapply (count_prop_mon (fun n => exists a, n = f a)) in LE.
    rewrite H in LE; simpl in LE; destruct LEM; [omega|eauto].  
Qed.

Lemma ci2_impl_ci: forall A
    (CNT: countably_infinite2 A),
  countably_infinite A.
Proof.
  i; rdes CNT; exists (compactify f); eapply inj_srj_impl_bij.
  - Case "Injectivity".
    eauto using compactify_inj.
  - Case "Surjectivity".
    red; intros; unfold compactify in *; induction b.
    + eapply compactify_le with (a:= g 0); omega.
    + des; edestruct (@nat_map_inj_impl_unbounded
                        (fun n => compactify f (g(n)))).
      * red; intros; eapply compactify_inj in H; eauto.
      * subst; eapply compactify_le; eauto.
Qed.

Lemma ci_impl_ci2: forall A
    (CNT: countably_infinite A),
  countably_infinite2 A.
Proof.
  intros; rdes CNT; exists f g; split; red; intros.
  - by rewrite <-(CNT x), <-(CNT y), H.
  - by rewrite <-(CNT0 x), <-(CNT0 y), H.
Qed.


(** ** Product preserves countable infinity. *)

Lemma cntinf_prod: forall A B
    (INFA: countably_infinite A)
    (INFB: countably_infinite B),
  countably_infinite (A * B).
Proof.
  i; apply ci2_impl_ci.
  red in INFA, INFB |- *; intros; des.
  exists (fun ab => nat_pair (f0 (fst ab), f (snd ab))).
  exists (fun n => (g0 n, g 0)).
  split; red; i.
  - destruct x, y; eapply nat_pair_inj in H; inv H.
    by rewrite <-(INFA a), <-(INFA a0), H1, <-(INFB b), <-(INFB b0), H2. 
  - by inv H; rewrite <-(INFA0 x), <-(INFA0 y), H1.
Qed.


(** ** [nat] is (trivially) countably infinite. *)

Theorem nat_cntinf:
  countably_infinite nat.
Proof.
  do 2 exists (@id nat); eauto.
Qed.
Hint Resolve nat_cntinf.



(** * Countably infinite sets. *)


(** ** Definition. *)

Definition countably_infinite_set A (N: A -> Prop) :=
  countably_infinite (sig N).


(** ** Splitting one countably infinite set into two. *)

Lemma countably_infinite_set_split: forall A (N: A -> Prop)
    (INF: countably_infinite_set N),
  exists N1 N2, 
    <<Nci1: countably_infinite_set N1>> /\
    <<Nci2: countably_infinite_set N2>> /\
    <<Nunion: N = nunion N1 N2>> /\
    <<Ndisj: ndisj N1 N2>>.
Proof.
  intros; do 2 red in INF; des.
  set (N1 a := exists i, a = proj1_sig (g (2 * i))).
  set (N2 a := exists i, a = proj1_sig (g (S(2 * i)))).
  set (f1 (a : sig N1) :=
    projT1 (constructive_indefinite_description _ (proj2_sig a))).
  set (f2 (a : sig N2) :=
    projT1 (constructive_indefinite_description _ (proj2_sig a))).
  set (g1 i := exist N1 (proj1_sig (g (2 * i))) (ex_intro _ _ eq_refl)).
  set (g2 i := exist N2 (proj1_sig (g (S(2 * i)))) (ex_intro _ _ eq_refl)).

  assert (F1F: forall a Pa1 Pa, f (exist _ a Pa) = 2 * (f1 (exist _ a Pa1))).
  {
    intros; unfold f1; gen_cid; intros (n & Pn); simpls.
    by rewrite <- INF0; f_equal; subst; eapply proj1_sig_inj.
  }

  assert (F2F: forall a Pa1 Pa, f (exist _ a Pa) = S(2 * (f2 (exist _ a Pa1)))).
  {
    intros; unfold f2; gen_cid; intros (n & Pn); simpls.
    by rewrite <- INF0; f_equal; subst; eapply proj1_sig_inj.
  }

  exists N1 N2; splits.
  - exists f1 g1; split; intros. 
    + destruct a as (a & Pa); unfold g1. 
      assert (Pa' : N a) by (rdes Pa; subst; apply (proj2_sig (g(2*i)))).
      gen_eqrefl; hrewrite <-(F1F _ Pa Pa'); rewrite INF.
      by simpls; eauto using proj1_sig_inj.
    + unfold f1, g1; gen_cid; intros (n' & Pn'); simpls.
      apply proj1_sig_inj in Pn'.
      cut (n+(n+0)=n'+(n'+0)); [by omega|].
      by rewrite <-INF0; rewrite <-INF0 at 1; rewrite Pn'.
  - exists f2 g2; split; intros.
    + destruct a as (a & Pa); unfold g2. 
      assert (Pa' : N a) by (rdes Pa; subst; apply (proj2_sig (g(S(2*i))))).
      gen_eqrefl; hrewrite <-(F2F _ Pa Pa'); rewrite INF.
      by simpls; eauto using proj1_sig_inj.      
    + unfold f2, g2; gen_cid; intros (n' & Pn'); simpls.
      apply proj1_sig_inj in Pn'.
      cut (S(n+(n+0))=S(n'+(n'+0))); [by omega|].
      by rewrite <-INF0; rewrite <-INF0 at 1; rewrite Pn'.
  - extensionality a; apply prop_ext; unfold nunion, N1, N2; split; intros X.
    + assert (Y := X); change a with (proj1_sig(exist _ _ X)) in Y.
      rewrite <- (INF (exist _ _ X)) in Y.
      destruct (even_or_odd (f (exist _ _ X))); des; [left|right]
      ; by exists x; rewrite <- H; rewrite INF.
    + des; subst; apply proj2_sig.
  - intros n (i & X) (j & Y); subst.
    eapply proj1_sig_inj in Y.
    cut (2*i = S(2*j)); [by omega|].
    by rewrite <- INF0; rewrite <- INF0 at 1; rewrite Y.
Qed.

Lemma countably_infinite_set_add: forall A (N: A -> Prop) el
    (INF: countably_infinite_set N),
  countably_infinite_set (fun a => N a \/ a = el).
Proof.
  intros; apply ci2_impl_ci.
  apply ci_impl_ci2 in INF; rdes INF.
  exists (fun a : {a | N a \/ a = el} =>
            let (a,PA) := a in
            match or2sum PA with
              | left PA => S (f (exist _ a PA))
              | right _ => 0
            end).
  exists (fun n => exist (fun a => N a \/ a = el)
            (proj1_sig (g n)) (or_introl _ (proj2_sig (g n)))).
  split; repeat intro.
  - destruct x as (x & PX), y as (y & PY); simpls.
    destruct (or2sum PX), (or2sum PY); inv H; eauto using proj1_sig_inj.
    by apply INF in H; inv H; eauto using proj1_sig_inj.
  - by inv H; eapply proj1_sig_inj, INF0 in H1.
Qed.

Lemma countably_infinite_set_remove: forall A (N: A -> Prop) el
    (INF: countably_infinite_set N),
  countably_infinite_set (fun a => N a /\ a <> el).
Proof.
  intros; destruct (LEM (N el)) as [NEL | NEL].

  - repeat red in INF; des. 
    set (nel := f (exist _ _ NEL)).

    assert (NOTEL: forall n, proj1_sig (g (S nel + n)) <> el).
    {
      red; intros; cut (S nel + n = nel); [by omega|].
      by rewrite <- INF0 at 1; unfold nel at 2; f_equal
      ; eauto using proj1_sig_inj.
    }

    apply ci2_impl_ci; repeat red in INF; des. 
    exists (fun a : sig (fun a => N a /\ a <> el) => 
              f (exist _ (proj1_sig a) (proj1 (proj2_sig a)))).
    exists (fun n => exist (fun a => N a /\ a <> el) _ 
              (conj (proj2_sig (g (S nel+n))) (NOTEL n))).
    split; red; intros.
    + eapply f_equal in H; rewrite !INF in H; inv H; eauto using proj1_sig_inj.
    + inv H; eapply proj1_sig_inj, f_equal in H1; rewrite !INF0 in H1; omega.

  - replace (fun a => N a /\ a <> el) with N; eauto.
    by extensionality a; apply prop_ext; intuition; subst.
Qed.

Lemma countably_infinite_set_split_finite: forall A (N: A -> Prop)
    (INF: countably_infinite_set N) L,
  exists N1 N2, 
    <<Nci1: countably_infinite_set N1>> /\
    <<Nci2: countably_infinite_set N2>> /\
    <<Nunion: N = nunion N1 N2>> /\
    <<Ndisj: ndisj N1 N2>> /\
    <<Ldisj: forall a (INL: List.In a L) (INN: N2 a), False>>.
Proof.
  induction L; i.
  - assert (X := countably_infinite_set_split INF); des; subst; eauto 10.
  - des; subst; unnw.
    exists (fun a' => N1 a' \/ (N2 a' /\ a' = a)).
    exists (fun a' => N2 a' /\ a' <> a).
    split.
    + destruct (LEM (N2 a)).
      * replace (fun a' => N1 a' \/ N2 a' /\ a' = a)
          with (fun a' => N1 a' \/ a' = a)
        ; [by eauto using countably_infinite_set_add|].
        extensionalities; apply prop_ext; split; i; des; subst; eauto.
      * replace (fun a' => N1 a' \/ N2 a' /\ a' = a) with N1; eauto.
        by extensionality a'; apply prop_ext; split; i; des; subst; eauto.
    + split; eauto using countably_infinite_set_remove; split.
      * extensionality x; apply prop_ext; unfold nunion; split
        ; i; destruct (LEM (x = a)); des; eauto.
      * split; ii; ss; des; eauto.
Qed.



(*********************************************************************************)

Class Enumerable (A: Type) :=
  { enum: A -> nat
  ; enum_inj: injective _ _ enum
  }.
Arguments enum _ _ _ : simpl never.


(** ** [nat] is enumerable. *)

Instance Enumerable_nat: Enumerable nat :=
  {| enum := id |}.
Proof.
  ii; auto.
Qed.


(** ** [A + B] is enumerable if A and B is. *)

Definition enum_sum A B (_: Enumerable A) (_: Enumerable B) : (A + B) -> nat :=
  fun x => match x with
             | inl a => nat_pair (0, enum a)
             | inr b => nat_pair (1, enum b)
           end.

Instance Enumerable_sum A (EA: Enumerable A) B (EB: Enumerable B) : Enumerable (A + B) :=
  {| enum := enum_sum EA EB |}.
Proof.
  ii; destruct x, y; f_equal
  ; apply nat_pair_inj in H; try apply EA; try apply EB; clarify.
Qed.


(** ** [A * B] is enumerable if A and B is. *)

Definition enum_prod A B (_: Enumerable A) (_: Enumerable B) : (A * B) -> nat :=
  fun x => nat_pair (enum (fst x), enum (snd x)).

Instance Enumerable_prod A (EA: Enumerable A) B (EB: Enumerable B) : Enumerable (A * B) :=
  {| enum := enum_prod EA EB |}.
Proof.
  ii; destruct x, y; f_equal
  ; apply nat_pair_inj in H; [apply EA|apply EB]; clarify.
Qed.


(** ** [list A] is enumerable if A is. *)

Fixpoint enum_list' A (enumA: A -> nat) (l: list A) : nat :=
  match l with
    | nil => 0
    | a :: l' => enum (enumA a, enum_list' enumA l')
  end.

Definition enum_list A (_: Enumerable A) : list A -> nat :=
  fun l => enum (length l, enum_list' enum l).

Instance Enumerable_list A (EA: Enumerable A) : Enumerable (list A) :=
  {| enum := enum_list EA |}.
Proof.
  unfold enum_list; ii; apply enum_inj in H; clarify.
  depgen y; induction x; destruct y; vauto; i.
  apply enum_inj in H; fold enum_list' in *; clarify.
  unfold id in H2; subst.            
  f_equal; eauto using enum_inj.
Qed.
