(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program.
Require Import vgtac basis.
Require Import lang_common lang_mid lang_tgt lang_spec.
Require Import model_common model_untyped.
Require Export gw_common.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



(** * References. *)

Record refst :=
  { refst_st :> location -> location -> Prop
  ; refst_bij: forall va va' vb vb'
        (IN: refst_st va vb)
        (IN': refst_st va' vb'),
      (va' = va -> vb' = vb) /\
      (vb' = vb -> va' = va)
  ; refst_fin: exists l,
      <<FIN: forall la lb (IN: refst_st la lb), In (la, lb) l>>
  }.
Arguments refst_bij [r va va' vb vb'] _ _.

Program Definition refst_emp: refst :=
  Build_refst (fun _ _ => False) _ _.
Next Obligation. vauto. Qed.
Next Obligation. ex @nil; vauto. Qed.

Program Definition ref_sts : transys :=
  {| state := refst
   ; ext s' s := s <r= s'
   ; extpub s' s := s <r= s'
  |}.

Definition refst_alloc (s: location -> location -> Prop) la lb
  : location -> location -> Prop :=
  s \2/ (fun la' lb' => la' = la /\ lb' = lb).

Program Definition refst_allocP (s: refst) la lb
    (NEWa: forall lb', ~ s la lb')
    (NEWb: forall la', ~ s la' lb) : refst :=
  Build_refst (refst_alloc s la lb) _ _.
Next Obligation.
  unfold refst_alloc in *; des; vauto.
  - hexploit (refst_bij IN IN'); eauto.
  - splits; i; subst; [hexploit NEWa|hexploit NEWb]; vauto.
  - splits; i; subst; [hexploit NEWa|hexploit NEWb]; vauto.
Qed.
Next Obligation.
  hexploit (refst_fin s); []; i; des.
  exists ((la, lb) :: l).
  ii; destruct IN; des; subst; s; eauto.
Qed.

Lemma refst_alloc_ext: forall la lb (s s': state ref_sts)
    (RST: refst_alloc s la lb = s'),
  ext ref_sts s' s.
Proof.
  destruct s'; ii; s.
  unfold refst_alloc in RST; ss; subst; eauto.
Qed.


(** Subtraction for [refst].  Used in proof of vertical compositionality. *)
Program Definition refst_minus (sr sr': refst) : refst :=
  {| refst_st la lb := sr la lb /\ ~ sr' la lb |}.
Next Obligation.
  hexploit (refst_bij H1 H); eauto.
Qed.
Next Obligation.
  hexploit (refst_fin sr); []; i; des.
  exists l; ii; eapply FIN; des; eauto.
Qed.


(** Extensionality principle for [refst]. *)
Lemma refst_extensionality: forall (sr sr': refst)
    (E: forall la lb, refst_st sr la lb <-> refst_st sr' la lb),
  sr = sr'.
Proof.
  i; destruct sr, sr'; s in E.
  assert (refst_st0 = refst_st1).
  { extensionalities; apply prop_ext; auto. }
  subst; f_equal; auto using proof_irrelevance.
Qed.



(** * Mid <-> Mid. *)

Definition mm_gwsts :=
  tsprod (tsprod mid_gw_sts mid_gw_sts) ref_sts.

Definition mm_rquery (s: mm_gwsts.(state)) va vb :=
  exists la lb, <<Ea: va = IVloc la>> /\
                <<Eb: vb = IVloc lb>> /\
                <<REG: snd s la lb>>.

Lemma mm_rquery_mon: forall (s: mm_gwsts.(state)) va vb s'
    (IN: mm_rquery s va vb)
    (EXT: ext _ s' s),
  mm_rquery s' va vb.
Proof.
  i; destruct s as [[]], s' as [[]], EXT.
  rrdes IN; ss; des; subst; rr; esplits; ss; eauto.
Qed.

Program Definition mm_gwquery : method_query mid_lsi mid_lsi mm_gwsts :=
  {| vquerya s := mid_vquery (fst (fst s))
   ; vqueryb s := mid_vquery (fst (fst s))
   ; cquerya s := mid_cquery (fst (fst s))
   ; cqueryb s := mid_cquery (fst (fst s))
   ; rquery := mm_rquery
  |}.
Next Obligation. inv WV; vauto. Qed.
Next Obligation. inv WV; vauto. Qed.
Next Obligation. inv CQ; eauto. Qed.
Next Obligation. inv CQ; eauto. Qed.
Next Obligation. eapply mm_rquery_mon; ss; eauto. Qed.

Definition mm_consconfrel (s: mm_gwsts.(state)) (ha: mid_lsi.(conf)) (hb: mid_lsi.(conf)) :=
  <<CPa: mid_confpred (fst (fst s)) ha>> /\
  <<CPb: mid_confpred (snd (fst s)) hb>>.


(** ** Configuration relation for references. *)

Definition mm_refconfrel' (sr: refst) g (s: mm_gwsts.(state)) (ha: mid_lsi.(conf)) (hb: mid_lsi.(conf)) :=
  exists ha' hb',
    <<Ea: ha = (ha', Nothing)>> /\
    <<Eb: hb = (hb', Nothing)>> /\
    <<VLDa: pmvalid ha'>> /\
    <<VLDb: pmvalid hb'>> /\
    <<DOMa: forall la v (ALC: pmread ha' la v), exists lb, sr la lb>> /\
    <<DOMb: forall lb v (ALC: pmread hb' lb v), exists la, sr la lb>> /\
    <<RELV: forall la lb (REG: sr la lb), exists va vb,
              <<RDa: pmread ha' la va>> /\
              <<RDb: pmread hb' lb vb>> /\
              <<RVL: vclos mm_gwquery s (g s) va vb>> >>.

Definition mm_refconfrel g s :=
  mm_refconfrel' (snd s) g s.


Lemma mm_refconfrel'_mon: forall sr g s ha hb
    (CR: mm_refconfrel' sr g s ha hb)
    g' (LE: g <g= g'),
  mm_refconfrel' sr g' s ha hb.
Proof.
  i; rrdes CR; rr; esplits; eauto.
  i; hexploit RELV; eauto; []; i; des.
  esplits; eauto using vclos_mon.
Qed.    

Lemma mm_refconfrel'_smon: forall sr g s ha hb
    (RCR: mm_refconfrel' sr g s ha hb)
    s' (EXT: ext _ s' s)
    (MON: g s <r= g s'),
  mm_refconfrel' sr g s' ha hb.
Proof.
  i; rrdes RCR; rr; esplits; eauto.
  i; hexploit RELV; eauto; []; i; des.
  esplits; eauto using vclos_mon.
Qed.

Lemma mm_refconfrel'_ext: forall sr sr' g sg ca cb
    (RCR: mm_refconfrel' sr' g sg ca cb)
    (EXT: ext ref_sts sr' sr),
  crelplus (mm_refconfrel' sr g sg)
           (mm_refconfrel' (refst_minus sr' sr) g sg) ca cb.
Proof.
  i; rrdes RCR; subst.
  hexploit (pmplus_from_pred ha' (fun la => exists lb, sr la lb)); []; i; des; subst.
  hexploit (pmplus_from_pred hb' (fun lb => exists la, sr la lb)); []; i; des; subst.
  dup VLDa; split_pmvalid VLDa; des.
  dup VLDb; split_pmvalid VLDb; des.
  exists ((hP, Nothing) : conf mid_lsi).
  exists ((hNP, Nothing) : conf mid_lsi).
  exists ((hP0, Nothing) : conf mid_lsi).
  exists ((hNP0, Nothing) : conf mid_lsi). 
  - rr; esplits; s; noemp; eauto.
    + rr; esplits; eauto; i.
      * apply RDP in ALC; des; eauto.
      * apply RDP0 in ALC; des; eauto.
      * dup REG; apply EXT, RELV in REG; des; esplits; eauto.
        { eapply (proj2 (RDP _ _)); eauto. }
        { eapply (proj2 (RDP0 _ _)); eauto. }
    + rr; esplits; eauto; i.
      * apply RDNP in ALC; des.
        apply DOMa in ALC; des; econs; rr; eauto.
      * apply RDNP0 in ALC; des.
        apply DOMb in ALC; des; econs; rr; eauto.
      * rrdes REG; dup REG; apply RELV in REG; des; esplits; eauto.
        { apply RDNP; split; eauto.
          i; rrdes FH; dup FH; apply EXT in FH.
          hexploit (refst_bij FH REG1); eauto; []; i; des.
          specialize (X eq_refl); des; subst; eauto.
        }
        { apply RDNP0; split; eauto.
          i; rrdes FH; dup FH; apply EXT in FH.
          hexploit (refst_bij FH REG1); eauto; []; i; des.
          specialize (X0 eq_refl); des; subst; eauto.
        }
Qed.

Lemma mm_refconfrel_mon: forall g s ha hb
    (CR: mm_refconfrel g s ha hb)
    g' (LE: g <g= g'),
  mm_refconfrel g' s ha hb.
Proof.
  i; eapply mm_refconfrel'_mon; eauto.
Qed.    

Lemma mm_refconfrel_fin: forall g s ha hb
    (RCR: mm_refconfrel g s ha hb),
  <<FINa: pmfinite (fst ha)>> /\ <<FINb: pmfinite (fst hb)>>.
Proof.
  i; hexploit (refst_fin (snd s)); []; i; des.
  rrdes RCR; subst; split.
  - exists (map fst l); i.
    apply DOMa in RD; des; apply FIN in RD.
    apply in_map_iff; econs; inst_pairs; s; eauto.
  - exists (map snd l); i.
    apply DOMb in RD; des; apply FIN in RD.
    apply in_map_iff; econs; inst_pairs; s; eauto.
Qed.


Program Definition mm_gwconf: method_conf mid_lsi mid_lsi mm_gwsts :=
  {| confrel g s := crelplus (mm_consconfrel s) (mm_refconfrel g s) |}.
Next Obligation.
  rrdes PR; rr; esplits; eauto using mm_refconfrel_mon.
Qed.

Definition mm_gw: world_glob mid_lsi mid_lsi :=
  {| gwsts := mm_gwsts
   ; gwquery := mm_gwquery
   ; gwconf := mm_gwconf
  |}.


(** Subtraction for [refst] part.  Used in proof of vertical compositionality. *)

Definition mm_minus (sg sg': mm_gw.(state)) : mm_gw.(state) :=
  let '(sma, smb, sr) := sg in
  let '(_, _, sr') := sg' in
  (sma, smb, refst_minus sr sr').

Lemma mm_minus_ext: forall s s',
  ext mm_gw s (mm_minus s s').
Proof.
  unfold mm_minus; i; destruct s as [[]], s' as [[]].
  s; splits; auto; i; des; auto.
Qed.


(** Misc. *)

Lemma mm_refconfrel_emp: forall g (sg: state mm_gw)
    (NOREF: snd sg <2= bot2) ha hb,
  mm_refconfrel g sg ha hb <-> ha = cemp /\ hb = cemp.
Proof.
  simpl cemp; unfold mid_cemp; split; intro RCR.
  - rrdes RCR; subst; s.
    split; repeat f_equal; apply pmemp_from_read; eauto; i.
    + apply DOMa in FH; des; apply NOREF in FH; vauto.
    + apply DOMb in FH; des; apply NOREF in FH; vauto.
  - des; subst.
    destruct sg; rr; esplits; s; eauto; i.
    + apply pmread_empD in ALC; vauto.
    + apply pmread_empD in ALC; vauto.
    + exfalso; eapply NOREF; rr; eauto.
Qed.



(** * Tgt <-> Mid. *)

Definition tm_gwsts :=
  tsprod (tsprod tgt_gw_sts mid_gw_sts) ref_sts.

Definition tm_rquery (s: tm_gwsts.(state)) (va: tgt_lsi.(val)) vb :=
  exists lb, <<Eb: vb = IVloc lb>> /\
             <<REG: snd s va lb>>.

Lemma tm_rquery_mon: forall (s: tm_gwsts.(state)) va vb
    (IN: tm_rquery s va vb)
    s' (EXT: ext _ s' s),
  tm_rquery s' va vb.
Proof.
  i; destruct s as [[]], s' as [[]], EXT.
  rrdes IN; ss; des; subst; rr; esplits; ss; eauto.
Qed.

Program Definition tm_gwquery : method_query tgt_lsi mid_lsi tm_gwsts :=
  {| vquerya s := tgt_vquery (fst (fst s))
   ; vqueryb s := mid_vquery (snd (fst s))
   ; cquerya s := tgt_cquery (fst (fst s))
   ; cqueryb s := mid_cquery (snd (fst s))
   ; rquery := tm_rquery
  |}.
Next Obligation. eauto using tgt_vquery_mon. Qed.
Next Obligation. inv WV; vauto. Qed.
Next Obligation. inv CQ; r; s; eauto. Qed.
Next Obligation. inv CQ; eauto. Qed.
Next Obligation. eapply tm_rquery_mon; ss; eauto. Qed.

Definition tm_consconfrel (s: tm_gwsts.(state)) (ha: tgt_lsi.(conf)) (hb: mid_lsi.(conf)) :=
  <<CPa: tgt_confpred (fst (fst s)) ha>> /\
  <<CPb: mid_confpred (snd (fst s)) hb>>.


(** ** Configuration relation for references. *)

Definition tm_refconfrel' (sr: refst) g (s: tm_gwsts.(state)) (ha: tgt_lsi.(conf)) (hb: mid_lsi.(conf)) :=
  exists ha' hb',
    <<Ea: ha = {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := ha' |}>> /\
    <<Eb: hb = (hb', Nothing)>> /\
    <<VLDa: pmvalid ha'>> /\
    <<VLDb: pmvalid hb'>> /\
    <<DOMa: forall la v (ALC: pmread ha' la v), exists lb, sr la lb>> /\
    <<DOMb: forall lb v (ALC: pmread hb' lb v), exists la, sr la lb>> /\
    <<RELV: forall la lb (REG: sr la lb), exists va vb,
              <<RDa: pmread ha' la va>> /\
              <<RDb: pmread hb' lb vb>> /\
              <<RVL: vclos tm_gwquery s (g s) va vb>> >>.

Definition tm_refconfrel g s :=
  tm_refconfrel' (snd s) g s.


Lemma tm_refconfrel'_mon: forall sr g s ha hb
    (RCR: tm_refconfrel' sr g s ha hb)
    g' (LE: g <g= g'),
  tm_refconfrel' sr g' s ha hb.
Proof.
  i; rrdes RCR; rr; esplits; eauto.
  i; hexploit RELV; eauto; []; i; des.
  esplits; eauto using vclos_mon.
Qed.

Lemma tm_refconfrel'_smon: forall sr g s ha hb
    (RCR: tm_refconfrel' sr g s ha hb)
    s' (EXT: ext _ s' s)
    (MON: g s <r= g s'),
  tm_refconfrel' sr g s' ha hb.
Proof.
  i; rrdes RCR; rr; esplits; eauto.
  i; hexploit RELV; eauto; []; i; des.
  esplits; eauto using vclos_mon.
Qed.

Lemma tm_refconfrel'_ext: forall sr sr' g sg ca cb
    (RCR: tm_refconfrel' sr' g sg ca cb)
    (EXT: ext ref_sts sr' sr),
  crelplus (tm_refconfrel' sr g sg)
           (tm_refconfrel' (refst_minus sr' sr) g sg) ca cb.
Proof.
  i; rrdes RCR; subst.
  hexploit (pmplus_from_pred ha' (fun la => exists lb, sr la lb)); []; i; des; subst.
  hexploit (pmplus_from_pred hb' (fun lb => exists la, sr la lb)); []; i; des; subst.
  dup VLDa; split_pmvalid VLDa; des.
  dup VLDb; split_pmvalid VLDb; des.
  exists {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := hP |}.
  exists {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := hNP |}.
  exists ((hP0, Nothing) : conf mid_lsi).
  exists ((hNP0, Nothing) : conf mid_lsi). 
  - rr; esplits; s; noemp; eauto.
    + rr; esplits; eauto; i.
      * apply RDP in ALC; des; eauto.
      * apply RDP0 in ALC; des; eauto.
      * dup REG; apply EXT, RELV in REG; des; esplits; eauto.
        { eapply (proj2 (RDP _ _)); eauto. }
        { eapply (proj2 (RDP0 _ _)); eauto. }
    + rr; esplits; eauto; i.
      * apply RDNP in ALC; des.
        apply DOMa in ALC; des; econs; rr; eauto.
      * apply RDNP0 in ALC; des.
        apply DOMb in ALC; des; econs; rr; eauto.
      * rrdes REG; dup REG; apply RELV in REG; des; esplits; eauto.
        { apply RDNP; split; eauto.
          i; rrdes FH; dup FH; apply EXT in FH.
          hexploit (refst_bij FH REG1); eauto; []; i; des.
          specialize (X eq_refl); des; subst; eauto.
        }
        { apply RDNP0; split; eauto.
          i; rrdes FH; dup FH; apply EXT in FH.
          hexploit (refst_bij FH REG1); eauto; []; i; des.
          specialize (X0 eq_refl); des; subst; eauto.
        }
Qed.

Lemma tm_refconfrel_mon: forall g s ha hb
    (CR: tm_refconfrel g s ha hb)
    g' (LE: g <g= g'),
  tm_refconfrel g' s ha hb.
Proof.
  i; eapply tm_refconfrel'_mon; eauto.
Qed.    

Lemma tm_refconfrel_fin: forall g s ha hb
    (RCR: tm_refconfrel g s ha hb),
  <<FINa: pmfinite (chp ha)>> /\ <<FINb: pmfinite (fst hb)>>.
Proof.
  i; hexploit (refst_fin (snd s)); []; i; des.
  rrdes RCR; subst; split.
  - exists (map fst l); i.
    apply DOMa in RD; des; apply FIN in RD.
    apply in_map_iff; econs; inst_pairs; s; eauto.
  - exists (map snd l); i.
    apply DOMb in RD; des; apply FIN in RD.
    apply in_map_iff; econs; inst_pairs; s; eauto.
Qed.

Lemma tm_refconfrel_shape: forall s r ca cb
    (CR: tm_refconfrel s r ca cb),
  exists refha refhb,
    ca = {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := refha |} /\
    cb = (refhb, Nothing).
Proof.
  i; destruct ca, cb; rrdes CR; ss; subst; eauto.
Qed.


Program Definition tm_gwconf : method_conf tgt_lsi mid_lsi tm_gwsts :=
  {| confrel g s := crelplus (tm_consconfrel s) (tm_refconfrel g s) |}.
Next Obligation.
  rrdes PR; rr; esplits; eauto using tm_refconfrel_mon.
Qed.

Definition tm_gw : world_glob tgt_lsi mid_lsi :=
  {| gwsts := tm_gwsts
   ; gwquery := tm_gwquery
   ; gwconf := tm_gwconf
  |}.


(** Subtraction for [refst] part.  Used in proof of vertical compositionality. *)

Definition tm_minus (sg sg': tm_gw.(state)) : tm_gw.(state) :=
  let '(st, sm, sr) := sg in
  let '(_, _, sr') := sg' in
  (st, sm, refst_minus sr sr').

Lemma tm_minus_ext: forall s s',
  ext tm_gw s (tm_minus s s').
Proof.
  unfold tm_minus; i; destruct s as [[]], s' as [[]].
  s; splits; auto; i; des; auto.
Qed.


(** Misc. *)

Lemma tm_refconfrel_emp: forall g (sg: state tm_gw)
    (NOREF: snd sg <2= bot2) ha hb,
  tm_refconfrel g sg ha hb <-> ha = cemp /\ hb = cemp.
Proof.
  simpl cemp; unfold tgt_cemp, mid_cemp; split; intro RCR.
  - rrdes RCR; subst; s.
    split; repeat f_equal; apply pmemp_from_read; eauto; i.
    + apply DOMa in FH; des; apply NOREF in FH; vauto.
    + apply DOMb in FH; des; apply NOREF in FH; vauto.
  - des; subst.
    destruct sg; rr; esplits; s; eauto; i.
    + apply pmread_empD in ALC; vauto.
    + apply pmread_empD in ALC; vauto.
    + exfalso; eapply NOREF; rr; eauto.
Qed.
